/*
  ibm_vpd_render - Convert IBM VPD chunks to text.
  
  This file is part of the Linux lsvpd package.
  
  (C) Copyright IBM Corp. 2002, 2003, 2004
  
  Maintained by  Martin Schwenke <martins@au.ibm.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
 
  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  $Id: ibm_vpd_render.c,v 1.12 2004/05/03 03:42:05 martins Exp $
*/
  
#define _GNU_SOURCE

#include <stdio.h>
#include <unistd.h>
#include <ctype.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <netinet/in.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "pci_vpd_to_txt.h"
#include "growbuf.h"
#include "util.h"

static void
usage()
{
	fprintf(stderr, "usage: %s file output-directory\n", vpd_program_name);
	exit(1);
}

static bool
is_printable(const unsigned char * s)
{
	while (*s != '\0') {
		if (!isprint((int) *s))
			return false;
		s++;
	}
	return true;
}

static void
dump_field(FILE * ksfd, char * k, char * v)
{
	FILE * fd;
	char * f;
	short i;

	if (is_printable(k) && is_printable(v) && (v[0] != '\0')) {
		for (i = 0; ; i++) {
			if (i == 0)
				f = k;
			else
				asprintf(&f, "%s.%03d", k, i);
			if (0 != access(f, F_OK))
				break;
			if (f != k)
				free(f);
		}

		if (NULL != (fd = fopen(f, "w"))) {
			fputs(v, fd);
			fclose(fd);
			fprintf(ksfd, "%s ", f);
		} 

		if (f != k)
			free(f);
	}
}

static void
dump_txt_vpd(growbuf * txt, const char * d)
{
	unsigned char * t1;
	unsigned char * t2;
	unsigned char * t3;
	FILE * fd = NULL;

	char * prev = getcwd(NULL, 0);

	growbuf_iter * i = growbuf_iter_new(txt);

	if ((0 != mkdir(d, 0777)) || (0 != chdir(d))) {
		free(prev);
		return;
	}

	if (NULL == (fd = fopen(".lsvpd", "w")))
		goto done;

	dump_field(fd, "FC", "????????");

	while (growbuf_iter_have(i, 4)) {
		growbuf_iter_next(i);
		t1 = growbuf_iter_current(i);
		growbuf_iter_move(i, 2);
		t2 = growbuf_iter_current(i);
		*t2 = '\0';
		growbuf_iter_next(i);
		if (!growbuf_iter_more(i))
			break;

		t2 = growbuf_iter_current(i);
		if (NULL == (t3 = memchr(t2, '\n', growbuf_iter_remaining(i))))
			break;

		*t3 = '\0';
		dump_field(fd, t1, t2);
		growbuf_iter_move(i, (off_t) (((unsigned char *) t3) - t2) +1);
	}

 done:
	chdir(prev);
	free(prev);
	if (NULL != fd)
		fclose(fd);
}

void
ibm_vpd_render(growbuf * s, const char * node)
{
	growbuf_iter * i = growbuf_iter_new(s);
	short subnode = 0;
	char * l;

	asprintf(&l, "%s/linux,vpd", node);

	if (0 == mkdir(l, 0777)) {
		while (growbuf_iter_have(i, 4)) {
			/* 4 byte, BIG-endian length. */
			uint32_t len = 0;
			int n;
			for (n = 0; n < 4; n++) {
				len = (len << 8) +
					(uint32_t) (* growbuf_iter_current(i));
				growbuf_iter_next(i);
			}
			
			if (growbuf_iter_have(i, len)) {
				growbuf * txt;
				/* Grab PCI VPD data */
				growbuf * pci_vpd =
					growbuf_append(growbuf_new(len),
						       growbuf_iter_current(i),
						       len);
				growbuf_iter_move(i, len);
				
				/* Convert to text and then render */
				if (NULL != (txt = pci_vpd_to_txt(pci_vpd))) {
					char * d;
					asprintf(&d, "%s/%03d", l, subnode);
					/* Dumping from text is
					 * probably about as fast as
					 * originally parsing to a
					 * data structure and then
					 * dumping the data structure.
					 * I might rewrite all of this
					 * later if it matters. */
					dump_txt_vpd(txt, d);
					free(d);
					growbuf_free(txt);
				}
				
				growbuf_free(pci_vpd);
			} else {
				break;
			}
			subnode++;
		}
	}

	free(l);
	growbuf_iter_free(i);
}

int
main(int argc, char ** argv)
{
	growbuf * vpd;

	char * data;
	size_t len;
	int err;

	vpd_program_name = argv[0];

	if (argc != 3) {
		usage();
	}

	if (!file_read_contents(argv[1], &data, &len, &err)) {
		die2(strerror(err), argv[1]);
	}

	vpd = growbuf_append(growbuf_new(len), data, len);
	free(data);

	ibm_vpd_render(vpd, argv[2]);

	growbuf_free(vpd);

	return 0;
}
