/*
  pci_vpd_rom_grab - Grab PCI VPD via the expansion ROM.
  
  Copyright (c) 2002, 2003 IBM, Martin Schwenke <martins@au.ibm.com>
  
  This file is part of the Linux lsvpd package.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  $Id: pci_vpd_rom_grab.c,v 1.4 2003/10/02 00:33:12 martins Exp $
*/

/* For asprintf: */
#define _GNU_SOURCE

#include <sys/stat.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <unistd.h>

#include <sys/ioctl.h>
#include <sys/mman.h>


#include <netinet/in.h>
#include <linux/pci.h>

#include "growbuf.h"
#include "util.h"
#include "pci_vpd_gen_grab.h"

static void
usage()
{
	fprintf(stderr, "usage: %s file\n", vpd_program_name);
	exit(1);
}

struct vpd_rom_state{
	struct vpd_state s;
	char *           data;
	size_t           size;
	uint16_t         vpd_ptr;
	uint16_t         vpd_cur;
};

static struct vpd_rom_state *
vpd_rom_create(vpd_read_func rf, int fd, uint32_t rom_base, uint32_t rom_size)
{
	uint16_t pci_data_struct;

	struct vpd_rom_state * ret = malloc(sizeof(struct vpd_rom_state));
	if (NULL == ret)
		goto bye;

	/*
	 * We used to try mmap()ing /dev/mem, but that didn't work
	 * well because rom_base is often relative to a PCI bridge.
	 * The solution below depends on HAVE_PCI_MMAP.  Happily this
	 * is defined on most of the interesting architectures: i386,
	 * ia64, ppc, sparc64, ppc64 (in 2.4.19).
	 */

	if (-1 == ioctl(fd, PCIIOC_MMAP_IS_MEM))
		goto yikes;

	ret->data = mmap(0, rom_size, PROT_READ, MAP_PRIVATE, fd, rom_base);
	if ((void *) -1 == ret->data)
		goto yikes;
	
	/* For munmap. */
	ret->size = rom_size;

	/*
	 * Here we compare to 0xffff as well, since that seems to be
	 * what we sometimes get when the ROM is bogus...
	 */

	pci_data_struct = le16_to_cpu(* (uint16_t *) (&(ret->data[0x18])));
	if ((0 == pci_data_struct) || (0xffff == pci_data_struct)) {
		munmap(ret->data, ret->size);
		goto yikes;
	}

	ret->vpd_ptr = le16_to_cpu(* (uint16_t *)
				   (&(ret->data[pci_data_struct + 0x08])));
	if ((0 == ret->vpd_ptr) || (0xffff == ret->vpd_ptr)) {
		munmap(ret->data, ret->size);
		goto yikes;
	}

	ret->vpd_cur = 0;
	ret->s.rf = rf;
 bye:
	return ret;
	
 yikes:
	free(ret);
	ret = NULL;
	goto bye;
	
}

static void
vpd_rom_destroy(struct vpd_rom_state * s)
{
	munmap(s->data, s->size);
	free(s);
}

static ssize_t
vpd_rom_read(struct vpd_state * this, uint8_t * c_ptr, ssize_t count)
{
	struct vpd_rom_state * x = (struct vpd_rom_state * ) this;

	if (x->vpd_ptr + x->vpd_cur + count > x->size) {
		return -1;
	}

	memcpy(c_ptr, &(x->data[x->vpd_ptr + x->vpd_cur]), count);
	x->vpd_cur += count;

	return count;
}

struct rom_info {
	int      fd;
	uint8_t  config_pos;
	uint32_t rom_base;
	uint32_t rom_size;
	bool we_enabled_it;
};

static bool
find_rom_info(struct rom_info * x)
{
	uint8_t type;
	uint32_t tmp;

	bool status = true;

	x->config_pos = 0;
	x->rom_base   = 0;

	/* Determine the position of the ROM base address or 0. */
	if (1 == big_cool_pread(x->fd, &type, 1, PCI_HEADER_TYPE)) {
		type &= ~(1 << 7);
		switch (type) {
		case PCI_HEADER_TYPE_NORMAL:
			x->config_pos = PCI_ROM_ADDRESS;
			break;
		case PCI_HEADER_TYPE_BRIDGE:
			x->config_pos = PCI_ROM_ADDRESS1;
			break;
		case PCI_HEADER_TYPE_CARDBUS:
			/* Don't know much about this... */
			x->config_pos = 0;
			break;
		default:
			fprintf(stderr,
				"%s: warning - "
				"unknown PCI header type=%02x\n",
				vpd_program_name,
				type);
			status = false;
		}
	} else {
		status = false;
	}

	/* Determine the address assigned to the ROM. */
	if (status && (4 == big_cool_pread(x->fd, (uint8_t *) &tmp,
					   4, x->config_pos))) {
		x->rom_base = le32_to_cpu(tmp);
		
		/* Is the unmasked bit is 0 then we don't want it. */
		if (0 == (x->rom_base & PCI_ROM_ADDRESS_MASK)) {
			status = false;
		}
	} else {
		/* Healthy paranoia... */
		status = false;
	}

	/* Determine size of ROM. */
	if (status) {
		tmp = cpu_to_le32((uint32_t) PCI_ROM_ADDRESS_MASK);
		if ((4 == big_cool_pwrite(x->fd, (uint8_t *) &tmp,
					  4, x->config_pos)) &&
		    (4 == big_cool_pread(x->fd, (uint8_t *) &tmp,
					 4, x->config_pos))) {
			x->rom_size = (~ le32_to_cpu(tmp)) + 1;
		} else {
			status = false;
		}
	}
	/* Restore the base address, enabling ROM if necessary. */
	if (status) {
		/* If it wasn't enabled, set the enable bit... */
		if (0 == (x->rom_base & PCI_ROM_ADDRESS_ENABLE)) {
			x->rom_base |= PCI_ROM_ADDRESS_ENABLE;
			x->we_enabled_it = true;
		}
		/* Write it back. */
		tmp = cpu_to_le32(x->rom_base);
		if (4 != big_cool_pwrite(x->fd, (uint8_t *) &tmp,
					 4, x->config_pos)) {
			status = false;
			/* FIXME: Hmmm... how do we know? */
			x->we_enabled_it = false;
		}
	}

	return status;
}

static void
disable_rom(struct rom_info * x)
{
	uint32_t tmp;

	x->rom_base &= ~PCI_ROM_ADDRESS_ENABLE;
	tmp = cpu_to_le32(x->rom_base);
	big_cool_pwrite(x->fd, (uint8_t *) &tmp, 4, x->config_pos);
}

int
main(int argc, char ** argv)
{
	int ret = 0;

	struct rom_info info;

	growbuf * vpd = NULL;

	vpd_program_name = argv[0];

	if (2 != argc) {
		usage();
	}
	
	info.fd = open(argv[1], O_RDWR);
	if (-1 == info.fd) {
		die2(strerror(errno), argv[1]);
	}
	
	if (find_rom_info(&info)) {

		struct vpd_rom_state * s =
			vpd_rom_create(vpd_rom_read,
				       info.fd,
				       info.rom_base & PCI_ROM_ADDRESS_MASK,
				       info.rom_size);

		if (NULL != s) {
			growbuf * frag = grab_vpd((struct vpd_state *) s);
			vpd_rom_destroy(s);

			if (NULL != frag) {
				/* Push length onto VPD, as in ibm,vpd. */
				uint32_t length = htonl(growbuf_length(frag));
				vpd = growbuf_append(vpd,
						     (uint8_t *)&length,
						     sizeof(uint32_t));
				/* Append frag to VPD. */
				vpd = growbuf_append(vpd,
						     growbuf_data(frag),
						     growbuf_length(frag));
				growbuf_free(frag);
			} else {
				ret = 1;
			}

		} else {			

			ret = 1;
		}
		
		if (info.we_enabled_it) {
			disable_rom(&info);
		}
	} else {
		ret = 1;
	}
	
	close(info.fd);

	if (NULL != vpd) {
		fwrite(growbuf_data(vpd), growbuf_length(vpd), 1, stdout);
		growbuf_free(vpd);
	} else {
		ret = 1;
	}

	return ret;
}
