/*
  pci_vpd_to_txt - Convert PCI VPD chunks to text.
  
  This file is part of the Linux lsvpd package.
  
  (C) Copyright IBM Corp. 2002, 2003, 2004
  
  Maintained by  Martin Schwenke <martins@au.ibm.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
 
  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  $Id: pci_vpd_to_txt.c,v 1.17 2004/02/16 03:40:20 martins Exp $
*/
  
#include <assert.h>
#include <stdlib.h>
#include <string.h>

#include "growbuf.h"
#include "pci_vpd_to_txt.h"
#include "util.h"

static uint8_t
checksum_data(uint8_t init, uint8_t * data, size_t len)
{
	size_t i;
	for (i = 0; i < len ; i++) {
		init += data[i];
	}
	return init;
}

static uint8_t
parse_uchar(growbuf_iter * i)
{
	uint8_t ret;

	assert(growbuf_iter_have(i, 1));
	ret = growbuf_iter_current(i)[0];
	growbuf_iter_next(i);

	return ret;
}

static bool
parse_uchar_len(growbuf_iter * i, size_t * len)
{
	if (growbuf_iter_more(i)) {
		*len = (size_t) growbuf_iter_current(i)[0];
		growbuf_iter_next(i);
		return true;
	} else {
		return false;
	}
}

static bool
parse_uleshort_len(growbuf_iter * i, size_t * len)
{
	if (growbuf_iter_have(i, 2)) {
		*len = (size_t) (* growbuf_iter_current(i));
		growbuf_iter_next(i);
		*len += (size_t) (256 * (* growbuf_iter_current(i)));
		growbuf_iter_next(i);	
		return true;
	} else {
		return false;
	}
}

static bool
parse_uchar_length_data(growbuf_iter * i, growbuf ** o)
{
	size_t s;
	if (parse_uchar_len(i, &s)) {
		if (growbuf_iter_have(i, s)) {
			*o = growbuf_append(*o, growbuf_iter_current(i), s);
			growbuf_iter_move(i,s);
			return true;
		} else {
			return false;
		}
	} else {
		return false;
	}
}

static bool
parse_uleshort_length_data(growbuf_iter * i, growbuf ** o)
{
	size_t s;
	if (parse_uleshort_len(i, &s)) {
		if (growbuf_iter_have(i, s)) {
			*o = growbuf_append(*o, growbuf_iter_current(i), s);
			growbuf_iter_move(i,s);
		}
		return true;
	} else {
		return false;
	}
}

static bool
parse_label(growbuf_iter * i, growbuf ** o)
{
	if (growbuf_iter_have(i,2)) {
		*o = growbuf_append(*o, growbuf_iter_current(i), 2);
		growbuf_iter_move(i,2);
	}

	return true;
}	

static growbuf *
parse_key_val(growbuf_iter * i)
{
	growbuf * kv = growbuf_new(128);

	kv = growbuf_append(kv, "*", 1);
	if (parse_label(i, &kv)) {
		kv = growbuf_append(kv, " ", 1);
		if (!parse_uchar_length_data(i, &kv)) {
			growbuf_free(kv);
			kv = NULL;
		}
	} else {
		growbuf_free(kv);
		kv = NULL;
	}

	return kv;
}
	
static void
parse_board_id(growbuf_iter * i, growbuf ** l, uint8_t * cs)
{
	growbuf * name = growbuf_new(128);
	char * ds_hdr = "*DS ";
	char * invalid = "NOT_VALID_VPD\n";
	uint8_t t;
	size_t len;
	char * yip;

	*l = growbuf_append(*l, ds_hdr, strlen(ds_hdr));
	
	if (!parse_uleshort_length_data(i, &name)) {
		/* This won't happen very often. */
		growbuf_free(name);
		name = growbuf_new(strlen(invalid));
		name = growbuf_append(name, invalid, strlen(invalid));
	}

	t = (uint8_t) growbuf_length(name);
	*cs = checksum_data(*cs, &t, sizeof(t));
	*cs = checksum_data(*cs, growbuf_data(name), growbuf_length(name));

	/* This works around a NUL character in name.  Yuck, huh? */
	len = growbuf_length(name);
	if (NULL != (yip = memchr(growbuf_data(name), '\0', len))) {
		len = (size_t) ((unsigned char *) yip - growbuf_data(name));
	}
	*l = growbuf_append(*l, growbuf_data(name), len);
	*l = growbuf_append(*l, "\n", 1);

	growbuf_free(name);
}

static void
parse_resource_chunk(growbuf_iter * i, growbuf ** l, uint8_t * cs)
{
	size_t len;
	uint8_t * start;

	/* Some VPD chunk lengths are broken, but we want to try to
	 * trust the VPD, so we actually advance by the key-value
	 * pairs until we've exceeded the stated length...  but only
	 * if we're not going to walk off the end of the earth!
	 */
	if (parse_uleshort_len(i, &len)) {
		size_t len2;
		growbuf * kv;
		*cs = checksum_data(*cs, (uint8_t *)&len, sizeof(len));

		start = growbuf_iter_current(i);

		while ((growbuf_iter_current(i) < (start + len)) &&
		       (NULL != (kv = parse_key_val(i)))) {

			/* This works around a NUL character in kv. */
			size_t len = growbuf_length(kv);
			char * yip =
				memchr(growbuf_data(kv), '\0', len);
			if (NULL != yip) {
				len = (size_t) ((unsigned char *) yip -
						growbuf_data(kv));
			}

			
			*l = growbuf_append(*l, growbuf_data(kv), len);
			*l = growbuf_append(*l, "\n", 1);

			growbuf_free(kv);
		}

		len2 = growbuf_iter_current(i) - start;
		if (len != len2) {
			fprintf(stderr,
				"%s: stated length of chunk (0x%04zx) != "
				"actual length (0x%04zx).\n",
				vpd_program_name, len, len2);
		}
		
		*cs = checksum_data(*cs, start, len2);
	}
}

/* This only generates warnings. */
static void
do_checksum(growbuf_iter * i, uint8_t cs, size_t cs_len)
{
	switch (cs_len) {
	case 0:
		break;
	case 1:
		/* We can only handle a 1 byte checksum.  This is what
		 * is used in ibm,vpd files.  I've never seen another
		 * example.  */
		if (growbuf_iter_more(i)) {
			cs += parse_uchar(i);
			if (cs != 0) {
				fprintf(stderr,
					"%s: checksum=0x%02x "
					"incorrect for resource.\n",
					vpd_program_name, cs);
			}
		} else {
			fprintf(stderr,
				"%s: missing checksum data.\n",
				vpd_program_name);
		}
		break;
	default:
		fprintf(stderr,
			"%s: unknown checksum length=%zd.\n",
			vpd_program_name, cs_len);
	}
}

#define RTT_ID 0x02
#define RTT_RO 0x10
#define RTT_RW 0x11
#define RTT_END 0x0f

growbuf * pci_vpd_to_txt(growbuf * s)
{
	uint8_t tag, name;
	uint8_t cs = 0;
	
	growbuf_iter * i = growbuf_iter_new(s);
	growbuf * ret = growbuf_new(512);
	
	do {
		if (!growbuf_iter_more(i)) {
			return NULL;
		}
		tag = parse_uchar(i);
		
		if ((tag & (1 << 7)) != 0) {
			/* Large Resource Data Type tag, length follows. */
			cs += tag;
			name = tag & ~(1 << 7);
			
			switch (name) {
			case RTT_ID:
				parse_board_id(i, &ret, &cs);
				break;
			case RTT_RO:
			case RTT_RW:
				parse_resource_chunk(i, &ret, &cs);
				break;
			default:
				fprintf(stderr, 
					"%s: internal error: "
					"Large Resource Name"
					"0x%02x not handled.\n",
					vpd_program_name, name);
				exit(1);
			}
		} else {
			name = (tag >> 3) & 0x0f;
			
			if (RTT_END == name) {
				do_checksum(i, cs, tag & 0x7);
			} else {
				fprintf(stderr, 
					"%s: internal error: "
					"Small Resource Name"
					"0x%02x not handled.\n",
					vpd_program_name, name);
				exit(1);
			}
		}
		
	} while (RTT_END != name);

	growbuf_iter_free(i);
	
	return ret;
}

